#!/usr/bin/env python
#   -*- encoding:   UTF-8   -*-

"""
    'The Web is Dead' (TWID) is a game for PyWeek 13
    Copyright (C) 2011 D Haynes

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

    tundish@thuswise.org
"""

import sys
import logging
import random
from time import gmtime, strftime
from collections import namedtuple
from math import factorial as bang

from altgraph.Graph import Graph

from model import Model
from topography import POPULATION, GEOGRAPHY, TOPOLOGY

NodeData = namedtuple("Data", ["model","static"])
LogEntry = namedtuple("Entry", ["time","gpid","event","msg"])

class Crisis(object):
    
    def __init__(self):
        self.graph = self._populate()
        self._seed = None
        self._t = 0.0
        self._log = []
 
    def _populate(self):
        g = Graph()
        faxNrs = set(["6%05d" % (i+1) for i in range(int(1E5))])
        for n, (k,popn) in enumerate(POPULATION.items()):
            if popn is None:
                continue
            model = Model(popn = popn)
            name = "%03d" % (n+1)
            places = GEOGRAPHY[k]
            faxNr = faxNrs.pop()
            data = NodeData(model, (name, places, faxNr))
            g.add_node(k,data)

        for k in g.node_list():
            for i in TOPOLOGY[k]:
                if i in g:
                    g.add_edge(k, i)

        return g

    @property
    def t(self):
        return self._t

    def fmtLog(self, tStart=0):
        return [LogEntry(strftime("%Y-%m-%d %H:%M:%SZ",gmtime(tStart+i.time)),
        self.nameOf(i.gpid),i.event,i.msg) for i in self._log]

    @property
    def q(self):
        n = self.graph.number_of_nodes()
        K = 2

        nC = bang(n)/(bang(K)*bang(n-K))
        return nC / float(n) # range 54 -> 0.5

    @property
    def seed(self):
        if self._seed is None:
            self._seed = random.choice(self.graph.node_list())
            self.graph.node_data(self._seed).model.start()
            self._log.append(LogEntry(
            self._t, self._seed, "genesis",
            "Unexplained phenomena in %s" % random.choice(
            self.placesIn(self._seed))))
        return self._seed

    def deploy(self, k):
        try:
            self.graph.node_data(k).model.set(3E-9, 9E-4)
        except KeyError:
            #FIXME: Overrun nodes are hidden. Data can't be retrieved
            return False
        else:
            state = self.graph.node_data(k).model.state
            self._log.append(LogEntry( state[0], k, "deployed",
            "Received order to deploy"))
            return True

    def nameOf(self, k):
        try:
            return self.graph.node_data(k).static[0]
        except KeyError:
            #FIXME: Overrun nodes are hidden. Data can't be retrieved
            return "???"

    def placesIn(self, k):
        return self.graph.node_data(k).static[1]

    def faxOf(self, k):
        return self.graph.node_data(k).static[2]

    def popnOf(self, k):
        return sum(self.graph.node_data(k).model.state[1:])

    def step(self, delta=None):
        rv = 0
        overrun = []
        for k in self.graph: # includes hidden nodes
            dt = self.graph.node_data(k).model.step(delta)
            rv = rv or dt
            
            if dt:
                state = self.graph.node_data(k).model.state
                level = float(state[2]) / sum(state[1:])
                if level > 0.61:
                    if k in self.graph.node_list():
                        self._log.append( LogEntry(
                        state[0], k, "overrun",
                        "Lost contact with node; mesh degraded"))
                        overrun.append(k)
                elif level > 0.1:
                    for i in self.graph.out_nbrs(k):
                        state = self.graph.node_data(i).model.state
                        if not state[3]:
                            self._log.append(LogEntry(
                            state[0], i, "infection",
                            "Suspected infection at %s" %
                            random.choice(self.placesIn(i))))
                            self.graph.node_data(i).model.start()

                else:
                    #TODO: reopen depot?
                    pass
    
        for k in overrun:
            self.graph.hide_node(k) # mutates graph

        if rv:
            self._t += rv

        return rv

