#!/usr/bin/env python
#   -*- encoding:   UTF-8   -*-

"""
    'The Web is Dead' (TWID) is a game for PyWeek 13
    Copyright (C) 2011 D Haynes

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

    tundish@thuswise.org
"""

import sys
import os.path
THIS_DIR = os.path.dirname(__file__)

import time
import threading
import random
import json
import subprocess

import bottle
from crisis import Crisis

class Simulation(threading.Thread):

    def __init__(self, crisis):
        threading.Thread.__init__(self)
        self.daemon = True
        self._tStart = 0
        self._time = None
        self._lock = threading.Lock()
        self._crisis = crisis

    @property
    def q(self):
        self._lock.acquire()
        rv = self._crisis.q
        self._lock.release()
        return rv

    @property
    def t(self):
        self._lock.acquire()
        rv = self._crisis.t
        self._lock.release()
        return rv

    @property
    def time(self):
        self._lock.acquire()
        rv = self._time
        self._lock.release()
        return rv

    @property
    def log(self):
        self._lock.acquire()
        rv = self._crisis.fmtLog(self._tStart)
        self._lock.release()
        return rv

    def deploy(self, k):
        self._lock.acquire()
        rv = self._crisis.deploy(k)
        self._lock.release()
        return rv

    def getIdFromName(self, name):
        self._lock.acquire()
        gids = [ i for i in self._crisis.graph
        if self._crisis.nameOf(i) == name]
        self._lock.release()

        if not len(gids):
            return None
        else:
            return gids[0]

    def getIdFromFax(self, fax):
        self._lock.acquire()
        gids = [ i for i in self._crisis.graph
        if self._crisis.faxOf(i) == fax]
        self._lock.release()

        if not len(gids):
            return None
        else:
            return gids[0]

    def nameOf(self, k):
        return self._crisis.nameOf(k)

    def placesIn(self, k):
        return self._crisis.placesIn(k)

    def faxOf(self, k):
        return self._crisis.faxOf(k)

    def popnOf(self, k):
        return self._crisis.popnOf(k)

    def run(self):
        self._lock.acquire()
        self._crisis.seed # accessor prompts seeding
        self._lock.release()
        while True:
            if not self._tStart:
                self._tStart = time.time()
                self._time = self._tStart
            else:
                self._lock.acquire()
                self._time += self._crisis.step()
                self._lock.release()
            time.sleep(0.1) #TODO: adjust

app = bottle.Bottle()
crisis = Crisis()
app.sim = Simulation(crisis)

@app.route('/views/:filename')
def server_static(filename):
    if len(sys.argv) > 1:
        viewsDir = sys.argv[1]
    else:
        viewsDir = os.path.join(THIS_DIR,"views")
    return bottle.static_file(filename, root=viewsDir)

@app.route("/log")
@bottle.view("log_page.tpl")
def log_page():
    # http://localhost:8080/log?n=30
    nItems = int(bottle.request.GET.get("n") or "30")
    return {"log":app.sim.log[-nItems:]}

@app.route("/score",method="GET")
def score_data():
    nItems = int(bottle.request.GET.get("n") or "30")
    return {"t":app.sim.t, "q":app.sim.q}

@app.route("/orders",method="PUT")
def orders_data():
    contents = bottle.request.body.read()
    data = json.loads(contents)
    fax = data.get("fax","")
    gid = app.sim.getIdFromFax(fax)
    if gid is None:
        bottle.abort(400, "Fax number not recognised")
    if not app.sim.deploy(gid):
        bottle.abort(404, "Resource not found.")
    return {"t":app.sim.t, "q":app.sim.q}

@app.route("/group/:grp#[0-9]+#")
@bottle.view("group_page.tpl")
def group_page(grp=0):
    name = "%03d" % int(grp)
    gid = app.sim.getIdFromName(name)
    if gid is None:
        bottle.abort(404, "Resource not found.")
    return {"name":name, "fax":app.sim.faxOf(gid),
    "popn":app.sim.popnOf(gid), "places":app.sim.placesIn(gid)}

def main():
    app.sim.start()
    if "debug" in sys.argv[-1]:
        debug = True
    else:
        debug = False
    bottle.debug(debug)
    bottle.run(app, host="localhost", port=8080, reloader=debug)

def run(fP, debug=False):
    """
    Permits invocation from Ren'Py interpreter. We find the path
    to this module with 'renpy.loader.transfn("server.py")' and
    then pass it in as the first argument. Now we can invoke the
    module as a subprocess and declare to bottle where to find
    our templates.
    """
    thisDir = os.path.dirname(fP)
    vDir = os.path.abspath(os.path.join(thisDir,"views"))
    proc = subprocess.Popen([sys.executable, fP, vDir, "deb_g"])
    proc.communicate() 

if __name__ == "__main__":
    if len(sys.argv) > 1:
        bottle.TEMPLATE_PATH.insert(0,sys.argv[1])
    main()

